@@ -22,7 +22,7 @@
 #
 # Bacula is a registered trademark of Kern Sibbald.
 
-__version__ = '0.1.6'
+__version__ = '0.1.7'
 
 import os
 import sys
@@ -37,6 +37,7 @@ import pexpect
 import fcntl
 import time
 import re
+import binascii
 
 from LogFile import *
 from Database import *
@@ -99,7 +100,18 @@ def touch(fname, times = None):
 
 class FileSystem(Fuse) :
 
-    null_stat = fuse.Stat(st_mode = stat.S_IFDIR | 0755, st_nlink = 2, st_ino = -1)
+    null_stat = fuse.Stat(st_mode=stat.S_IFDIR | 0755, 
+                          st_ino=0, 
+                          st_dev=0, 
+                          st_nlink=2, 
+                          st_uid=0, 
+                          st_gid=0, 
+                          st_size=0, 
+                          st_atime=0, 
+                          st_mtime=0, 
+                          st_ctime=0,
+                          st_blksize=0,
+                          st_rdev=0)
 
     bacula_stat_fields = ['st_dev',
                           'st_ino',
@@ -118,7 +130,7 @@ class FileSystem(Fuse) :
                           'st_flags',
                           'st_streamid']
 
-    fuse_stat_fields = dir(fuse.Stat())
+    fuse_stat_fields = [attr for attr in dir(null_stat) if attr.startswith('st_')]
 
     xattr_prefix = 'user.baculafs.'
     xattr_fields = ['FileIndex', 'JobId', 'LStat', 'MD5']
@@ -158,6 +170,7 @@ class FileSystem(Fuse) :
         self.datetime = None
         self.recent_job = False
         self.joblist = None
+        self.cache_prefix = None
         self.user_cache_path = None
         self.cleanup = False
         self.move_root = False
@@ -167,7 +180,14 @@ class FileSystem(Fuse) :
         self.prefetch_recent = False
         self.prefetch_diff = None
         self.prefetch_difflist = None
+        self.prefetch_list = None
         self.prefetch_everything = False
+        self.batch_mode = False
+        self.batch_list = False
+        self.batch_bsr = False
+        self.batch_extract = False
+        self.use_ino = False
+        self.max_ino = 0
         self.dirs = { '/': { '': (FileSystem.null_stat,) } }
 
         self._bextract_status = copy.deepcopy(FileSystem.bextract_done)
@@ -217,6 +237,19 @@ class FileSystem(Fuse) :
             self.dirs[head][tail] = (FileSystem.null_stat,)
         self._add_parent_dirs(head)
 
+    def _update_inodes(self, head) : 
+        '''
+        generate unique st_ino for each missing st_ino
+        '''
+        for tail in self.dirs[head] :
+            if self.dirs[head][tail][-1].st_ino == 0 :
+                if len(self.dirs[head][tail]) == 1:
+                    self.dirs[head][tail] = (copy.deepcopy(FileSystem.null_stat),)
+                self.max_ino += 1
+                self.dirs[head][tail][-1].st_ino = self.max_ino
+            subdir = '%s%s/' % (head, tail)
+            if subdir in self.dirs :
+                self._update_inodes(subdir)
     
     def _extract(self, path_list) :
         '''
@@ -256,7 +289,7 @@ class FileSystem(Fuse) :
             rc, sig = self._bextract(items)
             # it seems that bextract does not restore mtime for symlinks
             # so we create a normal file with same mtime as stored symlink
-            if rc == 0 :
+            if rc == 0 and not self.batch_mode :
                 for item in items :
                     if item[0] :
                         symlinkfile = item[0][0]
@@ -392,8 +425,31 @@ class FileSystem(Fuse) :
         '''
         extract list of items from Bacula storage device
         '''
+        if self.batch_list :
+            for item in items :
+                print item[1]
+            if (not self.batch_bsr and
+                not self.batch_extract) :
+                return (0, 0)
+
         bsrpath = self._write_bsr(items)
-        cmd = 'bextract -b "%s" -c "%s" "%s" "%s"' % (bsrpath, self.conf, self.device, self.cache_path)
+
+        if self.batch_bsr :
+            bsrfile = open(bsrpath, 'rt')
+            for line in bsrfile :
+                sys.stdout.write(line)
+            sys.stdout.flush()
+            bsrfile.close()
+            if not self.batch_extract :
+                return (0, 0)
+
+        if self.batch_extract :
+            makedirs(self.fuse_args.mountpoint)
+        
+        cmd = 'bextract -b "%s" -c "%s" "%s" "%s"' % (bsrpath,
+                                                      self.conf,
+                                                      self.device,
+                                                      self.cache_path if not self.batch_extract else self.fuse_args.mountpoint)
         self.logger.debug(cmd)
             
         self._bextract_set_status({'path': items[0][1],
@@ -513,7 +569,8 @@ class FileSystem(Fuse) :
             os.write(bsrfd, 'Device="%s"\n' % volume[2]) 
             os.write(bsrfd, 'VolSessionId=%d\n' % volume[4])
             os.write(bsrfd, 'VolSessionTime=%d\n' % volume[5])
-            os.write(bsrfd, 'VolAddr=%d-%d\n' % (volume[6],volume[7]))
+            if not self.bsr_compat :
+                os.write(bsrfd, 'VolAddr=%d-%d\n' % (volume[6],volume[7]))
             for findex in volume[8] :
                 if findex[0] == findex[1] :
                     os.write(bsrfd, 'FileIndex=%d\n' % findex[0])
@@ -560,14 +617,22 @@ class FileSystem(Fuse) :
                 self.logger.warning(traceback.format_exc())
         self.logfile = LogFile(self.logger, logging.DEBUG)
 
-    
-    def initialize(self):
+    def initialize(self, version) :
         '''
         initialize database, catalog
         '''
 
         self._setup_logging()
 
+        # batch mode
+        self.batch_mode = (self.batch_list or
+                           self.batch_bsr or
+                           self.batch_extract)
+        # disable INFO level logging in batch mode
+        if self.batch_mode and self.loglevel == logging.INFO :
+            self.loglevel = logging.WARNING
+            self.logger.setLevel(self.loglevel)
+        
         self.logger.info('Populating file system ... ')
 
         # setup cache
@@ -582,6 +647,10 @@ class FileSystem(Fuse) :
         self.cache_symlinks = os.path.normpath(self.cache_prefix + '/symlinks')
         makedirs(self.cache_symlinks)
 
+        # test for old version (2.x) of bacula
+        self.bsr_compat = int(version[0]) < 3
+        if self.bsr_compat :
+            self.logger.debug('Detected old Bacula: %s' % version)
         # test access to sd conf file
         open(self.conf, 'r').close()
         # init bextract failure pattren
@@ -613,6 +682,7 @@ class FileSystem(Fuse) :
             self.prefetch_regex = None
             self.prefetch_diff = None
             self.prefetch_difflist = None
+            self.prefetch_list = None
             self.prefetch_symlinks = True
         if self.prefetch_regex :
             try :
@@ -636,7 +706,7 @@ class FileSystem(Fuse) :
         if self.prefetch_difflist :
             self.prefetch_difflist = os.path.normpath(os.path.expanduser(self.prefetch_difflist))
             try :
-                difflistfile = open(self.prefetch_difflist, 'rt')
+                difflistfile = sys.stdin if self.prefetch_difflist == '-' else open(self.prefetch_difflist, 'rt')
                 for line in difflistfile.readlines():
                     date = ' '.join(line.split()[:5])
                     difflist[line[(len(date) + 1):].strip()] = time.strptime(date, '%a %b %d %H:%M:%S %Y')
@@ -646,10 +716,24 @@ class FileSystem(Fuse) :
                 # can't access/parse difflist: show traceback and ignore
                 self.logger.warning(traceback.format_exc())
                 self.prefetch_difflist = None
+        if self.prefetch_list :
+            self.prefetch_list = os.path.normpath(os.path.expanduser(self.prefetch_list))
+            try :
+                listfile = sys.stdin if self.prefetch_list == '-' else open(self.prefetch_list, 'rt')
+                matchlist = [line.strip() for line in listfile.readlines()]
+                listfile.close()
+                self.prefetch_symlinks = True
+            except :
+                # can't access/parse list: show traceback and ignore
+                self.logger.warning(traceback.format_exc())
+                self.prefetch_list = None
         if self.prefetch_recent :
             self.prefetch_symlinks = True
         if self.prefetch_symlinks :
             self.prefetch_attrs = True
+        if 'use_ino' in self.fuse_args.optlist:
+            self.use_ino = True
+            self.prefetch_attrs = True # must figure out max st_ino
 
         for file in files :
             head = file[0]
@@ -660,6 +744,10 @@ class FileSystem(Fuse) :
             # make file entry
             if self.prefetch_attrs :
                 entry = file[2:] + self._bacula_stat(file[-2])
+                # find max st_ino
+                if self.use_ino:
+                    if entry[-1].st_ino > self.max_ino :
+                        self.max_ino = entry[-1].st_ino
                 # detemine if we need to prefetch this entry
                 filepath = head + tail
                 if (not stat.S_ISDIR(entry[-1].st_mode) and
@@ -673,6 +761,8 @@ class FileSystem(Fuse) :
                      (self.prefetch_difflist and
                       (filepath[1:] not in difflist or 
                        difflist[filepath[1:]][:-1] != time.localtime(entry[-1].st_mtime)[:-1])) or 
+                     (self.prefetch_list and
+                      filepath in matchlist) or
                      (self.prefetch_symlinks and
                       stat.S_ISLNK(entry[-1].st_mode)))) :
                     prefetches.append(filepath)
@@ -689,6 +779,10 @@ class FileSystem(Fuse) :
             # and finally
             self.dirs[head][tail] = entry
 
+        # fix st_ino
+        if self.use_ino:
+            self._update_inodes('/')
+            
         npf = len(prefetches)
         if npf > 0 :
             self.logger.info('Prefetching %d objects ... ' % npf)
@@ -704,7 +798,7 @@ class FileSystem(Fuse) :
         '''
         remove cache directory if required
         '''
-        if self.cleanup and not self.user_cache_path :
+        if self.cleanup and not self.user_cache_path and self.cache_prefix :
             self.logger.info('removing cache directory: %s' % self.cache_prefix)
             shutil.rmtree(self.cache_prefix, ignore_errors = True)
         
@@ -742,6 +836,9 @@ class FileSystem(Fuse) :
             len(self.dirs[head][tail]) != 1 and
             n in FileSystem.xattr_fields) :
             val = str(self.dirs[head][tail][FileSystem.xattr_fields.index(n)])
+            if n == 'MD5' and val != '0':
+                l = len(val)
+                val = binascii.b2a_hex(binascii.a2b_base64(val+'='*((l*3+8)/3-l)+'\n')) # padding
         # attribute not found
         if val == None :
             return -errno.ENODATA
@@ -808,8 +905,11 @@ class FileSystem(Fuse) :
             yield fuse.Direntry(key)
         for key in self.dirs[path].keys() :
             if len(key) > 0:
-                bs = self.getattr(path + key)
-                ino = bs.st_ino if bs.st_ino != 0 else -1
+                if self.use_ino:
+                    bs = self.getattr(path + key)
+                    ino = bs.st_ino
+                else :
+                    ino = 0
                 yield fuse.Direntry(key, ino=ino)
             
     def readlink(self, path):
@@ -916,9 +1016,17 @@ BaculaFS: exposes the Bacula catalog and storage as a Filesystem in USErspace
     server.parser.add_option(mountopt="prefetch_diff", metavar="PATH", default=server.prefetch_diff,
                              help="extract files that do not match files at PATH (hint: speeds up rsync; implies prefetch_symlinks)")
     server.parser.add_option(mountopt="prefetch_difflist", metavar="DIFFLIST", default=server.prefetch_difflist,
-                             help="extract files that do not match files in DIFFLIST (list line format: 'Day Mon DD hh:mm:ss YYYY PATH'; hint: format matches output of 'duplicity list-current-files -v0 target_url'; implies prefetch_symlinks)")
+                             help="extract files that do not match files in DIFFLIST (list line format: 'Day Mon DD hh:mm:ss YYYY PATH'; use '-' to read from standard input; hint: format matches output of 'duplicity list-current-files -v0 target_url'; implies prefetch_symlinks)")
+    server.parser.add_option(mountopt="prefetch_list", metavar="LIST", default=server.prefetch_list,
+                             help="extract files that match files in LIST (list should contains one absolute file path per line; use '-' to read from standard input; implies prefetch_symlinks)")
     server.parser.add_option(mountopt="prefetch_everything", action="store_true", default=server.prefetch_everything,
                              help="extract everything upon filesystem initialization (complete restore to cache) [default: %default]")
+    server.parser.add_option(mountopt="batch_list", action="store_true", default=server.batch_list,
+                             help="list files to be prefetched and exit [default: %default]")
+    server.parser.add_option(mountopt="batch_bsr", action="store_true", default=server.batch_bsr,
+                             help="dump contnets of bsr file for extracting prefetched files and exit [default: %default]")
+    server.parser.add_option(mountopt="batch_extract", action="store_true", default=server.batch_extract,
+                             help="extract prefetched files to mount point and exit [default: %default]")
     server.parser.add_option(mountopt="user_cache_path", metavar="PATH", default=server.user_cache_path,
                              help="user specified cache path (hint: combine this with one of the prefetch options) [default: %default]")
     server.parser.add_option(mountopt="logging", choices=LOGGING_LEVELS.keys(), metavar='|'.join(LOGGING_LEVELS.keys()), default=server.logging,
@@ -935,12 +1043,13 @@ BaculaFS: exposes the Bacula catalog and storage as a Filesystem in USErspace
             # we initialize before main (i.e. not in fsinit) so that
             # any failure here aborts the mount
             try :
-                server.initialize()
+                server.initialize(bacula_version)
             except :
                 server.shutdown()
                 raise
 
-    server.main()
+    if not server.batch_mode :
+        server.main()
 
     # we shutdown after main, i.e. not in fsshutdown, because
     # calling fsshutdown with multithreaded==True seems to cause
