@@ -26,6 +26,6 @@
 """
 from FileSystem import __version__, main
 
-__all__ = ["FileSystem", "Database", "Catalog", "Base64", "SQL"] 
+__all__ = ["FileSystem", "Database", "Catalog", "Base64", "SQL", "LogFile"] 
 
 
